#include <iostream>
#include <cstdlib>
#include <cmath>
#include <ctime>
#include <string>
#include <vector>
#include <queue>
using namespace std;
// #include <GL/glut.h> // 環境によって変える
//#include "./glut.h"
#include <GLUT/glut.h> // for Mac
#pragma comment(lib, "./glut32.lib")

#define PI 3.141592
#define FULLSCREEN_MODE 0 // 0ならウィンドウ，1ならフルスクリーンでマウス非表示
#define HEADTRACKING_MODE 0 // OpenCVを使ったヘッドトラッキングを有効にするかどうか
#define WINDOW_WIDTH 640
#define WINDOW_HEIGHT 480
#define MAX_BALL 10000 // 画面上に表示する最大の玉の数
#define PLAYER 0 // 自機を表すtype
#define ENEMY 1 // 敵を表すtype

int AUTOPLAY = 1;
int key_up = 0, key_down = 0, key_left = 0, key_right = 0, key_space = 0; // キー状態を記録
int global_time; // 内部時計
int list; // 文字列表示用
int enemy_level; // 敵の強さのレベル
int is_game_start; // ゲーム開始・終了用
int sudden_stop; // 一時停止用
int enemySize = 1; // 最初の敵の数

#include "GLColor.h"
#include "PlayerTank.h"
#include "EnemyTank.h"
#include "Ball.h"
#include "Field.h"
#include "Parameter.h"

EnemyTank autoPlayer; // 自機
PlayerTank player; // AUTOPLAYモードの動的切り替えのために2つ用意

//EnemyTank enemyPool[MAX_ENEMY]; // 敵
vector<EnemyTank> enemyPool;

Field field;
Parameter parameter;

#if HEADTRACKING_MODE
#include <opencv2/objdetect/objdetect.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#define IMG_SCALE 2 // 1だと重すぎる，4だとちょっと離れたら認識しなくなる
using namespace cv;

queue<pair<float, pair<float, float> > > peekQue; // peek用キュー <radius, <x, y>>
float faceX = 0, faceY = 0, faceRadius = 1.0;
float myFaceX = WINDOW_WIDTH/2, myFaceY = WINDOW_HEIGHT/2 + 100;
// path to haarcascade_frontalface_alt.xml
String cascadeName = "/Users/usi3/Desktop/OpenCV-2.2.0/data/haarcascades/haarcascade_frontalface_alt.xml";
CvCapture* capture = 0;
Mat frame, frameCopy, image;  
CascadeClassifier cascade;
#endif


// ゲーム開始時にだけ必要な初期化を行う
void gameInit(void){
	srand(time(NULL));
    
	enemy_level = -1;
	is_game_start = 1;
	sudden_stop = 0;
}
// レベルアップ時にだけ必要な初期化を行う
void nextStageInit(void){
	enemy_level++;
	global_time = 0;
    
	autoPlayer.reset(PLAYER);
	player.reset(PLAYER);
    
	autoPlayer.maxLife *= enemySize;
	autoPlayer.life = player.maxLife = player.life = autoPlayer.maxLife;
    
    enemyPool.clear();
    for (int i=0; i<enemySize; i++) {
        // 勝手にresetされるはず
        enemyPool.push_back(*(new EnemyTank));
    }
    /*
     for (int i=0; i<enemySize; i++) {
     enemyPool[i].reset();
     }*/
}

// キー操作
void keyDown(unsigned char key, int x, int y){
	switch(key){
        case ' ':
            key_space = 1;
            break;
        case '\33':
            if(FULLSCREEN_MODE) glutLeaveGameMode();
            exit(0);
            break;
        case 'x':
            enemy_level++;
            break;
        case 'z':
            enemy_level--;
            break;
        case 'v':
            enemySize++;
            enemyPool.push_back(*(new EnemyTank));
            break;
        case 'c':
            if(enemyPool.size() > 1){
                enemySize--;
                enemyPool.pop_back();
            }
            break;            
        case 'r':
            is_game_start = 1;
            gameInit();
            nextStageInit();
            break;
        case 's':
            sudden_stop = sudden_stop ? 0 : 1;
            break;
		case 'm':
			if(AUTOPLAY){
				// playerにスイッチ
				AUTOPLAY = 0;
				player.setStatus(autoPlayer);
				autoPlayer.isExist = false;
				player.isExist = true;
			}else{
				// autoPlayerにスイッチ
				AUTOPLAY = 1;
				autoPlayer.setStatus(player);
				autoPlayer.isExist = true;
				player.isExist = false;
			}
            break;
        default:
            break;
	}
}
void keyUp(unsigned char key, int x, int y){
	switch(key){
        case ' ':
            key_space = 0;
            break;
        default:
            break;
	}
}
void specialKeyDown(int key, int x, int y){
	switch(key){
        case GLUT_KEY_UP:
            key_up = 1;
            break;
        case GLUT_KEY_DOWN:
            key_down = 1;
            break;
        case GLUT_KEY_LEFT:
            key_left = 1;
            break;
        case GLUT_KEY_RIGHT:
            key_right = 1;
            break;
        default:
            break;
	}
}
void specialKeyUp(int key, int x, int y){
	switch(key){
        case GLUT_KEY_UP:
            key_up = 0;
            break;
        case GLUT_KEY_DOWN:
            key_down = 0;
            break;
        case GLUT_KEY_LEFT:
            key_left = 0;
            break;
        case GLUT_KEY_RIGHT:
            key_right = 0;
            break;
        default:
            break;
	}
}

// 表示制御
void display(void){  
	if(sudden_stop) return;
	if(is_game_start){
		glClearColor(0, 0, 0, 0);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        
#if HEADTRACKING_MODE
        if (!peekQue.empty()) {
            pair<float, pair<float, float> > faceInfo = peekQue.back();
            peekQue.pop();
            faceX += (faceInfo.second.first/12 - faceX)*0.1;
            faceY += (faceInfo.second.second/6 - faceY)*0.1;
            faceRadius += ((faceInfo.first-30)/50 - faceRadius)*0.1;
            //cout << "faceY = " << faceY << endl;
            //cout << "faceRadius = " << faceRadius << endl;
        }
        
        // 視点変更
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluPerspective(32.0, (double)WINDOW_WIDTH / (double)WINDOW_HEIGHT, 1.0, 100.0);
        glTranslated(0.0, 0.0, 0.0);
        gluLookAt(0.0, 45.0, 35.0, faceX, 0.0, -faceY, 0.0, 1.0, 0.0);
#endif
        
        // オブジェクト描画
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
        
#if HEADTRACKING_MODE
        glScaled(faceRadius, faceRadius, faceRadius);
#endif

        
        // フィールド
        field.display();
		
		// 敵
        int enemyLife = 0;
        for (int i=0; i<enemyPool.size(); i++) {
            enemyPool[i].display();
            enemyLife += enemyPool[i].life;
        }  
        
		if(AUTOPLAY){
			// 自機
			autoPlayer.display();
			// パラメーター表示
			parameter.display(autoPlayer.life, enemyLife);
		}else{
			player.display();
			parameter.display(player.life, enemyLife);
		}
        
		// ダブルバッファリング
		glutSwapBuffers();
	}
	return;
}
// 画面 and 視点制御
void reshape(GLsizei w, GLsizei h){
	// 画面の大きさを変更する
    cout << "reshape w = " << w << "h = " << h << endl;
	glutReshapeWindow(WINDOW_WIDTH, WINDOW_HEIGHT);
    
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(32.0, (double)w / (double)h, 1.0, 100.0);
    
	glTranslated(0.0, 0.0, 0.0);
	gluLookAt(0.0, 45.0, 35.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
}

#if HEADTRACKING_MODE
void cvInit(){
    if(!cascade.load(cascadeName)){
        cerr << "ERROR: Could not load classifier cascade" << endl;
        return;
    }
    capture = cvCaptureFromCAM(0);
    if(!capture){
        cerr << "Capture from CAM didn't work" << endl;
        return;
    }
    cvNamedWindow("result", 1);
}

void detectAndDraw(Mat& img, CascadeClassifier& cascade){
    vector<Rect> faces;
    
    // CV_8UC1: 1チャネル8ビットunsigned char
    Mat gray, smallImg(cvRound(img.rows/IMG_SCALE), cvRound(img.cols/IMG_SCALE), CV_8UC1); 
    
    // 色空間をカラーからグレースケールへ変換
    cvtColor(img, gray, CV_BGR2GRAY);
    // IMG_SCALEに応じて小さくする
    resize(gray, smallImg, smallImg.size(), 0, 0, INTER_LINEAR);
    equalizeHist(smallImg, smallImg);
    
    double t = (double)cvGetTickCount();
    cascade.detectMultiScale( smallImg, faces,
                             1.1, 2, 0
                             //|CV_HAAR_FIND_BIGGEST_OBJECT
                             //|CV_HAAR_DO_ROUGH_SEARCH
                             |CV_HAAR_SCALE_IMAGE
                             ,
                             Size(30, 30) );
    t = (double)cvGetTickCount() - t;
    //printf("detection time = %g ms\n", t/((double)cvGetTickFrequency()*1000.));
    
    // 1つの顔にしか対応しないことにする
    vector<Rect>::const_iterator r = faces.begin();
    
    // 顔が1つ以上検出されているなら最初の顔のみについて処理
    if (r != faces.end()) {
        Point center;
        center.x = cvRound((r->x + r->width*0.5)*IMG_SCALE);
        center.y = cvRound((r->y + r->height*0.5)*IMG_SCALE);
        //cout << center.x << " " << center.y << endl;
        // ディスプレイを正面から見たとき顔は下の方に映るからyを半分引いてない．
        float radius = cvRound((r->width + r->height)*0.25*IMG_SCALE);
        circle( img, center, radius, CV_RGB(0,0,255), 3, 8, 0 );
        // radius  34 - 190 くらい
        peekQue.push(make_pair(radius, make_pair(center.x-myFaceX, center.y-myFaceY))); 
    }
    
    imshow("result", img);    
}

void cvTimer(){
    // for debug
    //if (rand()%100 < 10) {
    //int x = cos(global_time/5)*320;
    //int y = sin(global_time/5)*320;
    //peekQue.push(make_pair(x, y));
    //}
    
    IplImage* iplImg = cvQueryFrame(capture);
    frame = iplImg;
    if(frame.empty()){
        return;
    }
    if(iplImg->origin == IPL_ORIGIN_TL){
        frame.copyTo(frameCopy);
    }else{
        flip(frame, frameCopy, 0);
    }
    
    detectAndDraw(frameCopy, cascade);
}
#endif

// メインループ
void timer(int value){
	glutPostRedisplay();
	glutTimerFunc(50, timer, 0);
	if(sudden_stop) return;
    
#if HEADTRACKING_MODE
    cvTimer();
#endif
    
	if(is_game_start){
		global_time++;
        
        player.proc(enemyPool[0]);
		autoPlayer.proc(enemyPool[0]);
        
        // 自機と敵機が近づきすぎていたら離す
		if(AUTOPLAY){
			for (int i=0; i<enemyPool.size(); i++) {
				enemyPool[i].proc(autoPlayer);
			}
			for (int i=0; i<enemyPool.size(); i++) {
				if(autoPlayer.distance(enemyPool[i]) < 2.0){
					autoPlayer.move(-1.0);
					enemyPool[i].move(-1.0);
				}
			}
		}else{
			for (int i=0; i<enemyPool.size(); i++) {
				enemyPool[i].proc(player);
			}
			for (int i=0; i<enemyPool.size(); i++) {
				if(player.distance(enemyPool[i]) < 2.0){
					player.move(-1.0);
					enemyPool[i].move(-1.0);
				}
			}
		}
        
        
        // 敵機同士でも近づきすぎてたら離す
        for (int i=0; i<enemyPool.size(); i++) {
            for (int k=i+1; k<enemyPool.size(); k++) {
                if (enemyPool[i].distance(enemyPool[k]) < 2.0) {
                    enemyPool[i].move(0.5);
                    enemyPool[k].move(-0.5);
                }
            }
        }
        
		// 弾丸の当たり判定
		if(AUTOPLAY){
			for (int i=0; i<enemyPool.size(); i++) {
				if (autoPlayer.check(enemyPool[i])){
					// 敵の消滅処理
					enemyPool[i].attackable = false;
				}
			}
		}else{
			for (int i=0; i<enemyPool.size(); i++) {
				if (player.check(enemyPool[i])){
					enemyPool[i].attackable = false;
				}
			}
		}
        
        
        for (int i=0; i<enemyPool.size(); i++) {
			if(AUTOPLAY){
				enemyPool[i].check(autoPlayer);
			}else{
				enemyPool[i].check(player);
			}
        }
        
        // 敵同士の当たり判定も考える（総当り）
        for (int i=0; i<enemyPool.size(); i++) {
			if(!enemyPool[i].isExist) continue;
            for (int k=0; k<enemyPool.size(); k++) {
                if (i==k) {
                    continue;
                }
                if (enemyPool[i].check(enemyPool[k])) {
					// 敵の消滅処理
					enemyPool[k].attackable = false;
					//enemyPool[k].isExist = false;
                }
            }
        }
        
		// 最後にまとめて消す
		vector<EnemyTank>::iterator it;
		for(it=enemyPool.begin(); it!=enemyPool.end();){
			if((*it).isExist == false){
				it = enemyPool.erase(it);
			}else{
				it++;
			}
		}
        
        if (enemyPool.size() == 0) {
            nextStageInit();
			return;
        }
		if(AUTOPLAY){
			if(!autoPlayer.isExist && autoPlayer.expTimer==10){
				nextStageInit();
				return;
			}
		}
	}
    
	return;
}

void glInit(){
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
    
	if(FULLSCREEN_MODE){
		glutGameModeString("1024x768:32@60");
		glutEnterGameMode();
	}else{
		glutInitWindowPosition(100, 100);
		glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);
		glutCreateWindow("Tank Action! (OpenGL/GLUT)");
	}
    
	glutReshapeFunc(reshape);
	glutDisplayFunc(display);
	// キー操作
	glutKeyboardFunc(keyDown);
	glutKeyboardUpFunc(keyUp);
	glutSpecialFunc(specialKeyDown);
	glutSpecialUpFunc(specialKeyUp);
    
    glClearColor(0, 0, 0, 0);
    
	// 光源
    GLfloat light1Pos[] = {-16, 10, -16, 1};
    GLfloat light2Pos[] = {16, 10, -16, 1};
    GLfloat light3Pos[] = {16, 10, 16, 1};
    GLfloat light4Pos[] = {-16, 10, 16, 1};
    GLfloat lightCol[] = {1.0, 1.0, 1.0, 1};
    
	glLightfv(GL_LIGHT0, GL_POSITION, light1Pos);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, lightCol);
    glLightf(GL_LIGHT0, GL_SPECULAR, 1.0);
    //glLightf(GL_LIGHT0, GL_DIFFUSE, 0.7);
    glLightf(GL_LIGHT0, GL_AMBIENT, 0.2);
    
	glLightfv(GL_LIGHT1, GL_POSITION, light2Pos);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, lightCol);
    glLightf(GL_LIGHT1, GL_SPECULAR, 1.0);
    glLightf(GL_LIGHT1, GL_AMBIENT, 0.2);
    
    glLightfv(GL_LIGHT2, GL_POSITION, light3Pos);
	glLightfv(GL_LIGHT2, GL_DIFFUSE, lightCol);
    glLightf(GL_LIGHT2, GL_SPECULAR, 1.0);
    glLightf(GL_LIGHT2, GL_AMBIENT, 0.2);
    
    glLightfv(GL_LIGHT3, GL_POSITION, light4Pos);
	glLightfv(GL_LIGHT3, GL_DIFFUSE, lightCol);
    glLightf(GL_LIGHT3, GL_SPECULAR, 1.0);
    glLightf(GL_LIGHT3, GL_AMBIENT, 0.2);
    
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);
    glEnable(GL_LIGHT2);
	glEnable(GL_LIGHT3);
	glEnable(GL_DEPTH_TEST);
}

// 各種初期化とメインループへの移行
int main(int argc, char *argv[]){
	glutInit(&argc, argv);
    glInit();
    
#if HEADTRACKING_MODE
    cvInit();
#endif
    
	gameInit();
	nextStageInit();
	glutTimerFunc(50, timer, 0);
	glutMainLoop();
    
	return 0;
}
